# -*- coding: utf-8 -*-
"""
@author:XuMing(xuming624@qq.com)
@description: 
"""
import os

pwd_path = os.path.abspath(os.path.dirname(__file__))
# where to save custom files
extra_data_dir = os.path.join(pwd_path, "../extra_data")
if not os.path.exists(extra_data_dir):
    os.makedirs(extra_data_dir)

col_sep = ','  # separate label and content of train data
segment_sep = ' '  # separate cut word
train_file = os.path.join(extra_data_dir, "train.csv")
test_file = os.path.join(extra_data_dir, "test.csv")

sentence_delimiters = ['？', '！', '。', '；', '……', '…']
allow_speech_tags = ['an', 'i', 'j', 'l', 'n', 'nr', 'nrfg', 'ns', 'nt', 'nz', 't', 'v', 'vd', 'vn', 'eng']

# inner data file
stopwords_path = os.path.join(pwd_path, 'data/stopwords.txt')
person_name_path = os.path.join(pwd_path, 'data/person_name.txt')
place_name_path = os.path.join(pwd_path, 'data/place_name.txt')
common_char_path = os.path.join(pwd_path, 'data/common_char_set.txt')

# custom domain file for statistics
domain_sample_path = os.path.join(extra_data_dir, 'train.csv')
pmi_path = os.path.join(extra_data_dir, 'pmi_word_score.json')
entropy_path = os.path.join(extra_data_dir, 'entropy_word_score.json')
ngram = 4
model_path = os.path.join(extra_data_dir, 'classify_model.pkl')



# -*- coding: utf-8 -*-
"""
@author:XuMing(xuming624@qq.com)
@description: 
"""

from sklearn.ensemble import RandomForestClassifier
from sklearn.feature_extraction.text import TfidfVectorizer
from sklearn.metrics import classification_report
from sklearn.model_selection import train_test_split

from wordrank.feature import Feature
from wordrank.utils.io_utils import save_pkl, load_pkl
from wordrank.utils.logger import logger


def data_reader(file_path, col_sep='\t'):
    """
    Load data
    :param file_path:
    :param col_sep:
    :return: list, list: contents, labels
    """
    contents = []
    labels = []
    with open(file_path, 'r', encoding='utf-8') as f:
        for line in f:
            line = line.rstrip()
            parts = line.split(col_sep)
            if len(parts) != 2:
                continue
            contents.append(parts[0])
            labels.append(parts[1])
    return contents, labels


def tfidf_word_feature(data_set, is_infer=False, feature_vec_path='', word_vocab=None):
    """
    Get TFIDF ngram feature by word
    """
    if is_infer:
        vectorizer = load_pkl(feature_vec_path)
        data_feature = vectorizer.transform(data_set)
    else:
        vectorizer = TfidfVectorizer(analyzer='word', vocabulary=word_vocab, sublinear_tf=True)
        data_feature = vectorizer.fit_transform(data_set)
    vocab = vectorizer.vocabulary_
    logger.debug('vocab size: %d' % len(vocab))
    logger.debug(data_feature.shape)
    # if not self.is_infer:
    save_pkl(vectorizer, feature_vec_path, overwrite=True)
    return data_feature


def train(train_file,
          col_sep,
          stopwords_path,
          person_name_path,
          place_name_path,
          common_char_path,
          segment_sep,
          domain_sample_path,
          ngram,
          pmi_path,
          entropy_path,
          model_path,
          ):
    # 1.read train data
    contents, labels = data_reader(train_file, col_sep)
    logger.info('contents size:%s, labels size:%s' % (len(contents), len(labels)))

    # 2.get feature
    feat = Feature(stopwords_path=stopwords_path,
                   person_name_path=person_name_path,
                   place_name_path=place_name_path,
                   common_char_path=common_char_path,
                   segment_sep=segment_sep,
                   domain_sample_path=domain_sample_path,
                   ngram=ngram,
                   pmi_path=pmi_path,
                   entropy_path=entropy_path)

    features = []
    tags = []
    for content, label in zip(contents, labels):
        label_split = label.split(segment_sep)
        content_split = content.split(segment_sep)
        if len(label_split) != len(content_split):
            logger.warning('pass, content size not equal label size, %s %s' % (content, label))
            continue
        tags += label_split
        content_features, terms = feat.get_feature(content, is_word_segmented=True)
        features += content_features
    logger.info("[train]features size: %s, tags size: %s" % (len(features), len(tags)))
    assert len(features) == len(tags), "features size must equal tags size"
    X_train, X_val, y_train, y_val = train_test_split(features, tags, test_size=0.2, random_state=0)
    logger.debug("train size:%s, val size:%s" % (len(y_train), len(y_val)))
    # 3.train classification model, save model file
    model = RandomForestClassifier(n_estimators=300)
    # fit
    logger.debug("start train model ...")
    model.fit(X_train, y_train)
    # save model
    save_pkl(model, model_path, overwrite=True)
    logger.info("model saved: %s" % model_path)

    # 4.validation and evaluate
    logger.debug("evaluate model with validation data")
    evaluate(model, X_val, y_val)
    return model


def predict(query,
            model_path,
            stopwords_path,
            person_name_path,
            place_name_path,
            common_char_path,
            segment_sep,
            domain_sample_path,
            ngram,
            pmi_path,
            entropy_path,
            ):
    logger.info('model predict')
    # get feature
    feat = Feature(stopwords_path=stopwords_path,
                   person_name_path=person_name_path,
                   place_name_path=place_name_path,
                   common_char_path=common_char_path,
                   segment_sep=segment_sep,
                   domain_sample_path=domain_sample_path,
                   ngram=ngram,
                   pmi_path=pmi_path,
                   entropy_path=entropy_path
                   )
    features, terms = feat.get_feature(query, is_word_segmented=False)
    # predict classification model
    model = load_pkl(model_path)
    logger.debug("model predict")
    label_pred = model.predict(features)
    logger.info("words: %s" % terms)
    logger.info("predict label: %s" % label_pred)
    print("predict label: %s" % label_pred)
    return label_pred


def evaluate(model, test_data, test_label, pred_save_path=None):
    print('{0}, val mean acc:{1}'.format(model.__str__(), model.score(test_data, test_label)))
    # multi
    label_pred = model.predict(test_data)
    # precision_recall_curve: multiclass format is not supported
    print(classification_report(test_label, label_pred))
    if pred_save_path:
        with open(pred_save_path, 'w', encoding='utf-8') as f:
            for i in label_pred:
                f.write(str(i) + '\n')
    return label_pred



# -*- coding: utf-8 -*-
"""
@author:XuMing(xuming624@qq.com)
@description: cmd line
"""

import argparse

from wordrank import config
from wordrank.model import predict
from wordrank.model import train

if __name__ == '__main__':
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument('--train', action="store_true", default=False, help='do train or not')
    parser.add_argument('--train_file', type=str, default=config.train_file, help='train file, file encode utf-8.')
    parser.add_argument('--col_sep', type=str, default=config.col_sep, help='column sep')
    parser.add_argument('--segment_sep', type=str, default=config.segment_sep, help='segment_sep')
    parser.add_argument('--stopwords_path', type=str, default=config.stopwords_path, help='stop word file')
    parser.add_argument('--person_name_path', type=str, default=config.person_name_path, help='person name file')
    parser.add_argument('--place_name_path', type=str, default=config.place_name_path, help='place name file')
    parser.add_argument('--common_char_path', type=str, default=config.common_char_path, help='common_char_path')
    parser.add_argument('--domain_sample_path', type=str, default=config.domain_sample_path, help='domain_sample_path')
    parser.add_argument('--ngram', type=int, default=config.ngram, help='common_char_path')
    parser.add_argument('--pmi_path', type=str, default=config.pmi_path, help='pmi_path')
    parser.add_argument('--entropy_path', type=str, default=config.entropy_path, help='entropy_path')
    parser.add_argument('--model_path', type=str, default=config.model_path, help='model file path to save')
    parser.add_argument('--query', type=str, default='井冈山景点介绍', help='input query')
    args = parser.parse_args()
    print(args)
    if args.train:
        train(args.train_file,
              args.col_sep,
              args.stopwords_path,
              args.person_name_path,
              args.place_name_path,
              args.common_char_path,
              args.segment_sep,
              args.domain_sample_path,
              args.ngram,
              args.pmi_path,
              args.entropy_path,
              args.model_path)
    predict(args.query,
            args.model_path,
            args.stopwords_path,
            args.person_name_path,
            args.place_name_path,
            args.common_char_path,
            args.segment_sep,
            args.domain_sample_path,
            args.ngram,
            args.pmi_path,
            args.entropy_path)



# -*- coding: utf-8 -*-
"""
@author:XuMing(xuming624@qq.com)
@description:
"""
import json
import os
import pickle
from collections import defaultdict


def load_pkl(pkl_path):
    """
    加载词典文件
    :param pkl_path:
    :return:
    """
    with open(pkl_path, 'rb') as f:
        result = pickle.load(f)
    return result


def save_pkl(vocab, pkl_path, overwrite=True):
    """
    存储文件
    :param pkl_path:
    :param overwrite:
    :return:
    """
    if os.path.exists(pkl_path) and not overwrite:
        return
    with open(pkl_path, 'wb') as f:
        # pickle.dump(vocab, f, protocol=pickle.HIGHEST_PROTOCOL)
        pickle.dump(vocab, f, protocol=0)


def load_json(json_path, encoding='utf-8'):
    with open(json_path, mode='r', encoding=encoding) as json_file:
        data = json.load(json_file)
    return data


def save_json(data, json_path, mode='w', encoding='utf-8'):
    dir = os.path.dirname(os.path.abspath(json_path))
    if not os.path.exists(dir):
        print(dir)
        os.makedirs(dir)
    with open(json_path, mode=mode, encoding=encoding) as f:
        f.write(json.dumps(data, ensure_ascii=False, indent=4))


def write_vocab(vocab, filename):
    """Writes a vocab to a file

    Writes one word per line.

    Args:
        vocab: iterable that yields word
        filename: path to vocab file

    Returns:
        write a word per line

    """
    print("Writing vocab...")
    with open(filename, "w", encoding='utf-8') as f:
        for i, word in enumerate(vocab):
            if i != len(vocab) - 1:
                f.write(word + '\n')
            else:
                f.write(word)
    print("- write to {} done. {} tokens".format(filename, len(vocab)))


def load_vocab(filename):
    """Loads vocab from a file

    Args:
        filename: (string) the format of the file must be one word per line.

    Returns:
        d: dict[word] = index

    """
    try:
        d = dict()
        with open(filename, 'r', encoding='utf-8') as f:
            for idx, word in enumerate(f):
                word = word.strip()
                d[word] = idx

    except IOError:
        raise IOError(filename)
    return d


def build_vocab(items, sort=True, min_count=0, lower=False):
    """
    构建词典列表
    :param items: list  [item1, item2, ... ]
    :param sort: 是否按频率排序，否则按items排序
    :param min_count: 词典最小频次
    :param lower: 是否小写
    :return: list: word set
    """
    result = []
    if sort:
        # sort by count
        dic = defaultdict(int)
        for item in items:
            item = item if not lower else item.lower()
            dic[item] += 1
        # sort
        dic = sorted(dic.items(), key=lambda d: d[1], reverse=True)
        for i, item in enumerate(dic):
            key = item[0]
            if min_count and min_count > item[1]:
                continue
            result.append(key)
    else:
        # sort by items
        for i, item in enumerate(items):
            item = item if not lower else item.lower()
            result.append(item)
    return result



# -*- coding: utf-8 -*-
"""
@author:XuMing(xuming624@qq.com)
@description:

* 统计特征：包括PMI、IDF、textrank值、前后词互信息、左右邻熵、
独立检索占比（term单独作为query的qv/所有包含term的query的qv和）、统计概率TF、idf变种iqf
"""

from wordrank import config
from wordrank.features.pmi import PMI
from wordrank.features.text_feature import AttrDict
from wordrank.features.textrank import TextRank
from wordrank.features.tfidf import TFIDF
from wordrank.utils.tokenizer import word_segment


class StatisticsFeature(object):
    def __init__(self,
                 domain_sample_path=config.domain_sample_path,
                 ngram=4,
                 pmi_path=config.pmi_path,
                 entropy_path=config.entropy_path,
                 segment_sep=config.segment_sep
                 ):
        self.tfidf = TFIDF()
        self.text_rank = TextRank()
        self.segment_sep = segment_sep
        self.pmi_model = PMI(text=self.read_text(domain_sample_path),
                             ngram=ngram,
                             pmi_path=pmi_path,
                             entropy_path=entropy_path)

    def _get_tags_score(self, word, tags):
        score = 0.0
        for i, s in tags:
            if word == i:
                score = s
        return score

    @staticmethod
    def read_text(file_path, col_sep=config.col_sep, limit_len=100000):
        text = ''
        with open(file_path, 'r', encoding='utf-8') as f:
            for line in f:
                line = line.strip()
                parts = line.split(col_sep)
                text += parts[0]
        if limit_len > 0:
            result = text[:limit_len]
        else:
            result = text
        return result

    def get_feature(self, query, is_word_segmented=False):
        """
        Get statistics feature
        :param query:
        :param is_word_segmented:
        :return: list, list: term features, sentence features
        """
        term_features = []
        sentence_features = {}
        rank_tags = self.text_rank.textrank(query, withWeight=True)
        tfidf_tags = self.tfidf.extract_tags(query, withWeight=True)

        if is_word_segmented:
            word_seq = query.split(self.segment_sep)
        else:
            word_seq = word_segment(query, cut_type='word', pos=False)

        count = 0
        for word in word_seq:
            idf = self.tfidf.idf_freq.get(word, self.tfidf.median_idf)
            # PMI & Entropy
            left_word = word_seq[count - 1] if count > 0 else ''
            right_word = word_seq[count + 1] if count < len(word_seq) - 1 else ''
            left_right_word = ''.join([left_word, right_word])
            entropy_score = self.pmi_model.entropy_score(left_right_word)
            term_features.append(AttrDict(
                idf=idf,
                text_rank_score=self._get_tags_score(word, rank_tags),
                tfidf_score=self._get_tags_score(word, tfidf_tags),
                pmi_score=self.pmi_model.pmi_score(left_right_word),
                left_entropy_score=entropy_score[0],
                right_entropy_score=entropy_score[1],
            ))
            count += 1

        return term_features, sentence_features
